/*************************************************************************
 *
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * Copyright 2008 by Nakata, Maho
 * 
 * $Id: Cherk.debug.cpp,v 1.7 2009/09/16 00:41:44 nakatamaho Exp $ 
 *
 * MPACK - multiple precision arithmetic library
 *
 * This file is part of MPACK.
 *
 * MPACK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * MPACK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details
 * (a copy is included in the LICENSE file that accompanied this code).
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with MPACK.  If not, see
 * <http://www.gnu.org/licenses/lgpl.html>
 * for a copy of the LGPLv3 License.
 *
 ************************************************************************/
#include <mblas.h>
#include <blas.h>
#include <mpack_debug.h>

#if defined VERBOSE_TEST
#include <iostream>
#endif

#define MIN_K -2
#define MAX_K 8
#define MIN_N -2
#define MAX_N 8
#define MIN_LDA -2
#define MAX_LDA 8
#define MIN_LDC -2
#define MAX_LDC 8
#define MAX_ITER 3

void
Cherk_test3(const char *uplo, const char *trans, REAL alpha,
    REAL beta)
{
    int errorflag = FALSE;
    for (int n = MIN_N; n < MAX_N; n++) {
	for (int k = MIN_K; k < MAX_K; k++) {
	    int minlda;
	    if (Mlsame(trans, "N"))
		minlda = max(1, n);
	    else
		minlda = max(1, k);
	    for (int lda = minlda; lda < MAX_LDA; lda++) {
		for (int ldc = max(1, n); ldc < MAX_LDC; ldc++) {
#if defined VERBOSE_TEST
		  printf
		    ("#n is %d, k is %d, lda is %d, ldc is %d ",
		     n, k, lda, ldc);
		  printf("uplo is %s, trans is %s \n", uplo, trans);
#endif
		    COMPLEX *A;
		    COMPLEX *C;
		    complex < double >*Ad;
		    complex < double >*Cd;

		    int ka;

		    if (Mlsame(trans, "N"))
			ka = k;
		    else
			ka = n;

		    A = new COMPLEX[matlen(lda, ka)];
		    C = new COMPLEX[matlen(ldc, n)];
		    Ad = new complex < double >[matlen(lda, ka)];
		    Cd = new complex < double >[matlen(ldc, n)];

		    double alphad, betad;

		    alphad = cast2double(alpha);
		    betad = cast2double(beta);

		    for (int iter = 0; iter < MAX_ITER; iter++) {
			set_random_vector(A, Ad, matlen(lda, ka));
			set_random_vector(C, Cd, matlen(ldc, n));

		        mpack_errno = 0; blas_errno = 0;
			Cherk(uplo, trans, n, k, alpha, A, lda, beta, C, ldc);
			zherk_f77(uplo, trans, &n, &k, &alphad, Ad, &lda,
			    &betad, Cd, &ldc);
#if defined VERBOSE_TEST
			    printf("errno: mpack %d, blas %d\n", mpack_errno, blas_errno);
#endif
			     if (mpack_errno != blas_errno) {
			       printf("error in Mxerbla!!\n");
			       exit(1);
			     }

			double diff = infnorm(C, Cd, matlen(ldc, n), 1);
			if (fabs(diff) > EPSILON) {
			    printf("#error %.16e!\n", diff);
			    errorflag = TRUE;
			}
		    }
		    delete[]Cd;
		    delete[]Ad;
		    delete[]C;
		    delete[]A;
		}
	    }
	}
    }
    if (errorflag == TRUE) {
	printf("Cherk test failed...\n");
	exit(1);
    }
}

void
Cherk_test2(const char *side, const char *uplo)
{
    REAL alpha, beta;
    REAL Rzero, Rone;

    Rzero = 0.0;
    Rone = 1.0;

//a=*, b=*;
    alpha = mpf_randomnumber();
    beta = mpf_randomnumber();
    Cherk_test3(side, uplo, alpha, beta);

//a=*, b=1;
    alpha = mpf_randomnumber();
    beta = Rone;
    Cherk_test3(side, uplo, alpha, beta);

//a=1, b=*;
    alpha = Rone;
    beta = mpf_randomnumber();
    Cherk_test3(side, uplo, alpha, beta);

//a=*, b=0;
    alpha = mpf_randomnumber();
    beta = Rzero;
    Cherk_test3(side, uplo, alpha, beta);

//a=1, b=0;
    alpha = Rone;
    beta = Rzero;
    Cherk_test3(side, uplo, alpha, beta);

//a=1, b=1;
    alpha = Rone;
    beta = Rone;
    Cherk_test3(side, uplo, alpha, beta);

//a=0, b=*;
    alpha = Rzero;
    beta = mpf_randomnumber();
    Cherk_test3(side, uplo, alpha, beta);

//a=0, b=0;
    alpha = Rzero;
    beta = Rzero;
    Cherk_test3(side, uplo, alpha, beta);

//a=0, b=1;
    alpha = Rzero;
    beta = Rone;
    Cherk_test3(side, uplo, alpha, beta);

}

void
Cherk_test()
{
    Cherk_test2("U", "N");
    Cherk_test2("L", "N");
    Cherk_test2("U", "C");
    Cherk_test2("L", "C");
}

int
main(int argc, char *argv[])
{
#if defined ___MPACK_BUILD_WITH_GMP___
    int default_prec = 256;
    mpf_set_default_prec(default_prec);
    uniformrandom = new gmp_randclass(gmp_randinit_default);
#endif
#if defined ___MPACK_BUILD_WITH_QD___ || defined ___MPACK_BUILD_WITH_DD___
unsigned int oldcw;
fpu_fix_start(&oldcw);
#endif

    Cherk_test();
    printf("Cherk test passed...\n");

#if defined ___MPACK_BUILD_WITH_QD___ || defined ___MPACK_BUILD_WITH_DD___
fpu_fix_end(&oldcw);
#endif

  return(0);
}
