/*************************************************************************
 *
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * Copyright 2008 by Nakata, Maho
 * 
 * $Id: Rlamch.cpp,v 1.10 2009/11/24 02:23:56 nakatamaho Exp $ 
 *
 * MPACK - multiple precision arithmetic library
 *
 * This file is part of MPACK.
 *
 * MPACK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * MPACK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details
 * (a copy is included in the LICENSE file that accompanied this code).
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with MPACK.  If not, see
 * <http://www.gnu.org/licenses/lgpl.html>
 * for a copy of the LGPLv3 License.
 *
 ************************************************************************/

#include <mblas.h>
#include <mlapack.h>

#if defined ___MPACK_BUILD_WITH_GMP___
//"E" denots we always calculate relative machine precision (e).
//where 1+e > 1, minimum of e.
REAL RlamchE_gmp(void)
{
    static REAL eps;
    static int called = 0;
    if (called)
	return eps;
    REAL one;
    unsigned long exp2;
    one = 1.0;
    exp2 = mpf_get_prec(one.get_mpf_t());
    mpf_div_2exp(eps.get_mpf_t(), one.get_mpf_t(), exp2);
    called = 1;
    return eps;
}

//"S" denots we always calculate `safe minimum, such that 1/sfmin does not overflow'.
//cf.http://www.netlib.org/blas/dlamch.f
REAL RlamchS_gmp(void)
{
    REAL sfmin;
    REAL one = 1.0;
    unsigned long exp2;

    exp2 = (1UL << (mp_bits_per_limb-8)) - 1; //6 seems to be the smallest on amd64 but for safty
    mpf_div_2exp(sfmin.get_mpf_t(), one.get_mpf_t(), exp2);
    return sfmin;

/* following code fragment is to test safe minimum 
    REAL largenum;
    for(int p = 60; p>=0; p--) { 
      for (int a=16; a<= 5120; a=a+128) {
        sfmin = 0.0; 
	mpf_set_default_prec(a);
	exp2 = (1UL << (mp_bits_per_limb-p)) - 1;
	mpf_div_2exp(sfmin.get_mpf_t(), one.get_mpf_t(), exp2);
	largenum  = 1.0 / sfmin;
	gmp_printf("%d, a:%5d, p:%5d sfmin: %16.20Fe largenum: %16.20Fe\n", mp_bits_per_limb, a, p, sfmin.get_mpf_t(), largenum.get_mpf_t());
	if (sfmin < 1.0 ) { printf("sfmin yes\n"); }
	if (largenum > 1.0 ) { printf("largenum yes\n"); }
      }
    }
*/

}
//"B" base  = base of the machine
//cf.http://www.netlib.org/blas/dlamch.f
REAL RlamchB_gmp(void)
{
    REAL two;
    two = 2.0;
    return two;
}

//"P" prec = eps*base
//cf.http://www.netlib.org/blas/dlamch.f
REAL RlamchP_gmp(void)
{
    REAL base, eps, prec;

    base = RlamchB_gmp();
    eps = RlamchE_gmp();
    prec = eps * base;
    return prec;
}

//"N" t = number of digits in mantissa
//cf.http://www.netlib.org/blas/dlamch.f
REAL RlamchN_gmp(void)
{
    unsigned long int tmp;
    REAL mtmp;
    REAL mtmp2;
    tmp = mpf_get_prec(mtmp.get_mpf_t());
    mtmp2 = tmp;
    return mtmp2;

/* following is fragment of code to test digits in mantissa
   for (int a=8; a<= 5120; a=a+128) {
     mpf_set_default_prec(a);
     REAL mmtmp;
     tmp = mpf_get_prec(mmtmp.get_mpf_t());
     printf("number of digits in mantissa %d\n", (int)tmp );
   }
*/

}

//"R" rnd   = 1.0 when rounding occurs in addition, 0.0 otherwise
//cf.http://www.netlib.org/blas/dlamch.f
REAL RlamchR_gmp(void)
{
//always rounding in addition on GMP.
    REAL mtmp;

    mtmp = 1.0;
    return mtmp;
}

//"M"
//cf.http://www.netlib.org/blas/dlamch.f
REAL RlamchM_gmp(void)
{
    unsigned long exp2;
    REAL tmp;
    REAL uflowmin, one=1.0; 
    exp2 = (1UL << (mp_bits_per_limb-8)) - 1; //6 seems to be the smallest on amd64 but for safty
    tmp = exp2; 
    return -tmp;

/*
  following code fragment is to test minimum 
  exponent before (gradual) underflow...but we just got Bus error
    REAL mtmp1, mtmp2;
    for(int p = 11; p>=0; p--) { 
      for (int a=51200; a<= 102400; a=a+128) {
	mpf_set_default_prec(a);
	printf("p %d a %d \n", p, a);
        uflowmin=0.0;
	exp2 = (1UL << (mp_bits_per_limb-p)) - 1;
	mpf_div_2exp(uflowmin.get_mpf_t(), one.get_mpf_t(), exp2);
        mtmp1 = uflowmin/2.0;
	gmp_printf("p %d, uflowmin: %16.20Fe uflowmin/2 %16.20Fe\n", p, uflowmin.get_mpf_t(), mtmp1.get_mpf_t());
  	mtmp2 = mtmp1 * 2.0;
	gmp_printf("mtmp2: %16.20Fe %lu\n", mtmp2.get_mpf_t(),exp2);
        if (uflowmin != mtmp2 ) { printf("underflow\n"); exit(1); }
      }
    }
*/
}

//"U"
//cf.http://www.netlib.org/blas/dlamch.f
REAL RlamchU_gmp(void)
{
    REAL underflowmin;
    REAL one = 1.0;
    unsigned long exp2;
    exp2 = (1UL << (mp_bits_per_limb-8)) - 1; //6 seems to be the smallest on amd64 but for safty
    mpf_div_2exp(underflowmin.get_mpf_t(), one.get_mpf_t(), exp2);
    return underflowmin;
}

//"L"
//cf.http://www.netlib.org/blas/dlamch.f
REAL RlamchL_gmp(void)
{
    REAL maxexp;
    unsigned long exp2;
    exp2 = (1UL << (mp_bits_per_limb-8)) - 1; //6 seems to be the smallest on amd64 but for safty
    maxexp = exp2;
    return maxexp;
}

//"O"
//cf.http://www.netlib.org/blas/dlamch.f
REAL RlamchO_gmp(void)
{
    REAL overflowmax;
    REAL one = 1.0;
    unsigned long exp2;

    exp2 = (1UL << (mp_bits_per_limb-8)) - 1; //6 seems to be the smallest on amd64 but for safty
    mpf_mul_2exp(overflowmax.get_mpf_t(), one.get_mpf_t(), exp2);

    return overflowmax;
}

//"Z" :dummy
//cf.http://www.netlib.org/blas/dlamch.f
REAL RlamchZ_gmp(void)
{
    REAL mtemp = 0.0;
    return mtemp;
}

REAL Rlamch_gmp(const char *cmach)
{
    if (Mlsame(cmach, "E"))
	return RlamchE_gmp();
    if (Mlsame(cmach, "S"))
	return RlamchS_gmp();
    if (Mlsame(cmach, "B"))
	return RlamchB_gmp();
    if (Mlsame(cmach, "P"))
	return RlamchP_gmp();
    if (Mlsame(cmach, "N"))
	return RlamchN_gmp();
    if (Mlsame(cmach, "R"))
	return RlamchR_gmp();
    if (Mlsame(cmach, "M"))
	return RlamchM_gmp();
    if (Mlsame(cmach, "U"))
	return RlamchU_gmp();
    if (Mlsame(cmach, "L"))
	return RlamchL_gmp();
    if (Mlsame(cmach, "O"))
	return RlamchO_gmp();

    Mxerbla("Rlamch", 1);
    return RlamchZ_gmp();
}
#endif

#if defined ___MPACK_BUILD_WITH_QD___
//"E" denots we always calculate relative machine precision (e).
//where 1+e > 1, minimum of e.
qd_real RlamchE_qd(void)
{
  //2^-209 = 1.21e-63
  return qd_real::_eps;
}

//"S" denots we always calculate `safe minimum, such that 1/sfmin does not overflow'.
//cf.http://www.netlib.org/blas/dlamch.f
qd_real RlamchS_qd(void)
{
  //2^(-1022+3*53) = 1.626e-260
  return qd_real::_min_normalized;
}
//"B" base  = base of the machine
//cf.http://www.netlib.org/blas/dlamch.f
qd_real RlamchB_qd(void)
{
    qd_real two;
    two = 2.0;
    return two;
}

//"P" prec = eps*base
//cf.http://www.netlib.org/blas/dlamch.f
qd_real RlamchP_qd(void)
{
    qd_real base, eps, prec;

    base = RlamchB_qd();
    eps = RlamchE_qd();
    prec = eps * base;
    return prec;
}

//"N" t = number of digits in mantissa
//cf.http://www.netlib.org/blas/dlamch.f
qd_real RlamchN_qd(void)
{
  return (qd_real)208.0; //52*4
}

//"R" rnd   = 1.0 when rounding occurs in addition, 0.0 otherwise
//cf.http://www.netlib.org/blas/dlamch.f
qd_real RlamchR_qd(void)
{
    qd_real mtmp;
    mtmp = 1.0;
    return mtmp;
}

//"M"
//cf.http://www.netlib.org/blas/dlamch.f
qd_real RlamchM_qd(void)
{
  return qd_real(-1022.0 + 3.0*53.0);
}

//"U"
//cf.http://www.netlib.org/blas/dlamch.f
qd_real RlamchU_qd(void)
{
  return qd_real::_min_normalized;
}

//"L"
//cf.http://www.netlib.org/blas/dlamch.f
qd_real RlamchL_qd(void)
{
  return (qd_real)1024.0;
}

//"O"
//cf.http://www.netlib.org/blas/dlamch.f
qd_real RlamchO_qd(void)
{
    return qd_real::_max; //approx 1.7976931348623157E+308 in float.h
}

//"Z" :dummy
//cf.http://www.netlib.org/blas/dlamch.f
qd_real RlamchZ_qd(void)
{
    qd_real mtemp = 0.0;
    return mtemp;
}

qd_real Rlamch_qd(const char *cmach)
{
    if (Mlsame(cmach, "E"))
	return RlamchE_qd();
    if (Mlsame(cmach, "S"))
	return RlamchS_qd();
    if (Mlsame(cmach, "B"))
	return RlamchB_qd();
    if (Mlsame(cmach, "P"))
	return RlamchP_qd();
    if (Mlsame(cmach, "N"))
	return RlamchN_qd();
    if (Mlsame(cmach, "R"))
	return RlamchR_qd();
    if (Mlsame(cmach, "M"))
	return RlamchM_qd();
    if (Mlsame(cmach, "U"))
	return RlamchU_qd();
    if (Mlsame(cmach, "L"))
	return RlamchL_qd();
    if (Mlsame(cmach, "O"))
	return RlamchO_qd();

    Mxerbla("Rlamch", 1);
    return RlamchZ_qd();
}
#endif

#if defined ___MPACK_BUILD_WITH_DD___
//"E" denots we always calculate relative machine precision (e).
//where 1+e > 1, minimum of e.
dd_real RlamchE_dd(void)
{
  //2^(-52-52) = 2^-104 = 4.93e-32
  return dd_real::_eps;
}

//"S" denots we always calculate `safe minimum, such that 1/sfmin does not overflow'.
//cf.http://www.netlib.org/blas/dlamch.f
dd_real RlamchS_dd(void)
{
  //2^(-1022+53) = 2.0042e-292
  return dd_real::_min_normalized;
}
//"B" base  = base of the machine
//cf.http://www.netlib.org/blas/dlamch.f
dd_real RlamchB_dd(void)
{
    dd_real two;
    two = 2.0;
    return two;
}

//"P" prec = eps*base
//cf.http://www.netlib.org/blas/dlamch.f
dd_real RlamchP_dd(void)
{
    dd_real base, eps, prec;

    base = RlamchB_dd();
    eps = RlamchE_dd();
    prec = eps * base;
    return prec;
}

//"N" t = number of digits in mantissa
//cf.http://www.netlib.org/blas/dlamch.f
dd_real RlamchN_dd(void)
{
  return (dd_real)104.0; //52*2
}

//"R" rnd   = 1.0 when rounding occurs in addition, 0.0 otherwise
//cf.http://www.netlib.org/blas/dlamch.f
dd_real RlamchR_dd(void)
{
    dd_real mtmp;
    mtmp = 1.0;
    return mtmp;
}

//"M"
//cf.http://www.netlib.org/blas/dlamch.f
dd_real RlamchM_dd(void)
{
  return dd_real(-1022.0 + 53.0);
}

//"U"
//cf.http://www.netlib.org/blas/dlamch.f
dd_real RlamchU_dd(void)
{
  return dd_real::_min_normalized;
}

//"L"
//cf.http://www.netlib.org/blas/dlamch.f
dd_real RlamchL_dd(void)
{
  return (dd_real)1024.0;
}

//"O"
//cf.http://www.netlib.org/blas/dlamch.f
dd_real RlamchO_dd(void)
{
    return dd_real::_max; //approx 1.7976931348623157E+308 in float.h
}

//"Z" :dummy
//cf.http://www.netlib.org/blas/dlamch.f
dd_real RlamchZ_dd(void)
{
    dd_real mtemp = 0.0;
    return mtemp;
}

dd_real Rlamch_dd(const char *cmach)
{
    if (Mlsame(cmach, "E"))
	return RlamchE_dd();
    if (Mlsame(cmach, "S"))
	return RlamchS_dd();
    if (Mlsame(cmach, "B"))
	return RlamchB_dd();
    if (Mlsame(cmach, "P"))
	return RlamchP_dd();
    if (Mlsame(cmach, "N"))
	return RlamchN_dd();
    if (Mlsame(cmach, "R"))
	return RlamchR_dd();
    if (Mlsame(cmach, "M"))
	return RlamchM_dd();
    if (Mlsame(cmach, "U"))
	return RlamchU_dd();
    if (Mlsame(cmach, "L"))
	return RlamchL_dd();
    if (Mlsame(cmach, "O"))
	return RlamchO_dd();

    Mxerbla("Rlamch", 1);
    return RlamchZ_dd();
}
#endif

REAL
Rlamc3(REAL a, REAL b)
{
  return a + b;
}
